#ifndef lint
static const char rcsid[] =
  "$FreeBSD: src/usr.sbin/pkg_install/add/main.c,v 1.33 2001/02/17 20:04:41 obrien Exp $";
#endif

/*
 *
 * FreeBSD install - a package for the installation and maintainance
 * of non-core utilities.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 * Jordan K. Hubbard
 * 18 July 1993
 *
 * This is the add module.
 *
 */

#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include "list.h"
#include "lib.h"

typedef enum { NORMAL, MASTER, SLAVE, PORT } add_mode_t;

static char Options[] = "hvIRfnrp:SMt:OU:";

static void usage(void);
void cleanup(int);

int
main(int argc, char **argv)
{
	int ch, err = 0;
	char tmp[PATH_MAX];
	char *upgrade = NULL;
	add_mode_t mode = NORMAL;
	boolean remote = FALSE;

	signal(SIGINT, cleanup);
	signal(SIGHUP, cleanup);

	pkg_mgr_init();
	while ((ch = getopt(argc, argv, Options)) != -1) {
		switch(ch) {
		case 'v':
			PkgMgr.verbose = TRUE;
			break;

		case 'p':
			strlcpy(PkgMgr.prefix_override, optarg, PATH_MAX);
			break;

		case 'I':
			PkgMgr.noinstall = TRUE;
			break;

		case 'R':
			PkgMgr.norecord = TRUE;
			break;

		case 'f':
			PkgMgr.force = TRUE;
			break;

		case 'n':
			PkgMgr.fake = TRUE;
			PkgMgr.verbose = TRUE;
			break;

		case 'r':
			remote = TRUE;
			break;

		case 't':
			strlcpy(PkgMgr.mktmp_template, optarg, PATH_MAX);
			break;

		case 'S':
			mode = SLAVE;
			break;

		case 'M':
			mode = MASTER;
			break;

		case 'O':
			mode = PORT;
			break;
			
		case 'U':
			upgrade = optarg;
			break;
			
		case 'h':
		case '?':
		default:
			usage();
			break;
		}
	}
	argc -= optind;
	argv += optind;

	switch (mode) {
		case PORT:
			if (argc) {
				msg_warn("cannot supply package name(s) in port mode");
				usage();
			}
			else if (read_line(tmp, PATH_MAX, stdin) == NULL) {
				msg_warn("pkg_add in PORT mode not passed a directory");
				return FAIL;
			}
			if (pkg_mgr_add_port(tmp)) {
				msg_warn("Port mode package addition failed");
				return FAIL;
			}
			break;
			
		case SLAVE:
			if (argc) {
				msg_warn("cannot supply package name(s) in slave mode");
				usage();
			}
			else if (read_line(tmp, PATH_MAX, stdin) == NULL) {
				msg_warn("pkg_add in SLAVE mode not passed a directory");
				return FAIL;
			}
			if (pkg_mgr_add_slave(tmp)) {
				msg_warn("Slave mode package addition failed");
				return FAIL;
			}
			break;

		case MASTER:
			if (argc != 1) {
				msg_warn("only one package name must be specified with master mode");
				usage();
			} 
			if (pkg_mgr_add_master(*argv, tmp) || strlen(tmp) == 0) {
				msg_warn("Master mode package extraction failed");
				return FAIL;
			}
			fprintf(stdout, tmp);
			break;
			
		case NORMAL:
			/* If no packages, yelp */
			if (!argc) {
				msg_warn("missing package name(s)");
				usage();
			}
			while (*argv) {
				if (remote)
					err += pkg_mgr_add_latest(*argv);
				else
					err += pkg_mgr_add(*argv);
				argv++;
			}
			if (err != 0) {
				msg_warn("%d package addition(s) failed", err);
				return err;
			}
	}
	return SUCCESS;
}

static void
usage()
{
	fprintf(stderr, "%s\n%s\n",
				"usage: pkg_add [-vInrfRMSO] [-t template] [-p prefix]",
				"			   pkg-name [pkg-name ...]");
	exit(1);
}

void
cleanup(int sig)
{
	static int in_cleanup = 0;

	if (!in_cleanup) {
		in_cleanup = 1;
			printf("Signal %d received, cleaning up..\n", sig);
	}
	if (sig)
		exit(1);
}
