/*
 * FreeBSD install - a package for the installation and maintainance
 * of non-core utilities.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 * Jordan K. Hubbard
 * 18 July 1993
 *
 * This is the info module.
 *
 */

#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include "list.h"
#include "lib.h"

#ifndef lint
static const char rcsid[] =
  "$FreeBSD: src/usr.sbin/pkg_install/info/main.c,v 1.31 2001/03/23 18:45:24 sobomax Exp $";
#endif

static char Options[] = "I:l:thvV";

static void usage __P((void));
void cleanup(int);

int
main(int argc, char **argv)
{
	int ch, err = 0;
	pkg_entry *pe;

	signal(SIGINT, cleanup);
	signal(SIGHUP, cleanup);

	if (argc == 4 && !strcmp(argv[1], "-t")) {
		err = version_cmp(argv[2], argv[3]);
		printf(err > 0 ? ">\n" : (err < 0 ? "<\n" : "=\n"));
		return 0;
	}
	
	pkg_mgr_init();
	while ((ch = getopt(argc, argv, Options)) != -1) {
		switch(ch) {
		case 'v':
			PkgMgr.verbose = TRUE;
			break;

		case 'l':
			strlcpy(PkgIndex.mask, optarg, 7);
			break;

		case 'I':
			strlcpy(PkgIndex.index, optarg, PATH_MAX);
			break;

		case 'V':
			break;
					
		case 't':
		case 'h':
		default:
			usage();
			break;
		}
	}

	argc -= optind;
	argv += optind;

	/* Get all the remaining package names, if any */
	if (*argv == NULL) {
		PLIST_FOREACH(pe, &PkgMgr) {
			err += pkg_mgr_version(pe->dbdir);
		}
	}
	else {
		while (*argv) {
			err += pkg_mgr_version(*argv);
			argv++;
		}
	}
	if (err != 0) {
		msg_warn("%d package(s) failed", err);
	}
	
	return err;
}

static void
usage()
{
	fprintf(stderr, "%s\n%s\n",
		"usage: pkg_admin -V [-v] [-l mask] -[I index] [packages]",
		"	    pkg_admin -t v1 v2");
	exit(1);
}

void
cleanup(int sig)
{
	static int in_cleanup = 0;

	if (!in_cleanup) {
		in_cleanup = 1;
			printf("Signal %d received, cleaning up..\n", sig);
	}
	if (sig)
		exit(1);
}
