#ifndef lint
static const char rcsid[] =
  "$FreeBSD: src/usr.sbin/pkg_install/create/main.c,v 1.25 2001/01/24 08:22:53 sobomax Exp $";
#endif

/*
 * FreeBSD install - a package for the installation and maintainance
 * of non-core utilities.
 *
 * Jordan K. Hubbard
 * 18 July 1993
 *
 * This is the create module.
 *
 */

#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include "list.h"
#include "lib.h"

typedef enum { NORMAL, BACKUP, PORT } create_mode_t;

static char Options[] = "OBhvf:p:P:c:d:i:I:k:K:r:t:X:D:m:s:L:R:o:";

static void usage __P((void));
void cleanup(int);

int
main(int argc, char **argv)
{
	int ch, err = 0;
	char *cp;
	char tmp[PATH_MAX];
	create_mode_t mode = NORMAL;
	pkg_info pi = { PKG_INFO_INITIALIZER };
	depend_list dl = { { PLIST_HEAD_INITIALIZER }, PLIST_HEAD_INITIALIZER };
	pkg_entry *pe;

	signal(SIGINT, cleanup);
	signal(SIGHUP, cleanup);

	pkg_mgr_init();
	while ((ch = getopt(argc, argv, Options)) != -1) {
		switch(ch) {
		case 'v':
			PkgMgr.verbose = TRUE;
			break;

		case 'O':
			mode = PORT;
			break;

		case 'B':
			mode = BACKUP;
			break;

		case 'p':
			pi.prefix = optarg;
			break;

		case 's':
			pi.srcdir = optarg;
			break;

		case 'f':
			pi.contents = optarg;
			break;

		case 'c':
			pi.comment = optarg;
			break;

		case 'd':
			pi.desc = optarg;
			break;

		case 'i':
			pi.install = optarg;
			break;

		case 'k':
			pi.deinstall = optarg;
			break;

		case 'r':
			pi.require = optarg;
			break;

		case 't':
			strlcpy(PkgMgr.mktmp_template, optarg, PATH_MAX);
			break;

		case 'X':
			PkgMgr.excludefrom = optarg;
			break;

		case 'h':
			PkgMgr.dereference = TRUE;
			break;

		case 'D':
			pi.display = optarg;
			break;

		case 'm':
			pi.mtree = optarg;
			break;

		case 'P':
			while (optarg) {
				cp = strsep(&optarg, " \t\n");
				add_depend_entry(&dl, DEPEND_PKG, NULL, cp);
			}
			break;

		case 'L':
			if ((cp = strchr(optarg, ':')))
				*cp++ = '\0';
			else
				cp = NULL;
			add_depend_entry(&dl, DEPEND_LIB, optarg, cp);
			break;

		case 'R':
			if ((cp = strchr(optarg, ':')))
				*cp++ = '\0';
			else
				cp = NULL;
			add_depend_entry(&dl, DEPEND_RUN, optarg, cp);
			break;

		case 'o':
			pi.origin = optarg;
			break;

		case '?':
		default:
			usage();
			break;
		}
	}
	argc -= optind;
	argv += optind;

	switch (mode) {
		case PORT:
			if (argc != 1) {
				msg_warn("only one package name can be specified for creatation");
				usage();
			} 
			PkgMgr.verbose = FALSE;
			if (pkg_mgr_create_port(&pi, &dl, *argv, tmp) || strlen(tmp) == 0) {
				msg_warn("Port mode package creation failed");
				return FAIL;
			}
			fprintf(stdout, "%s\n", tmp);
			break;
			
		case BACKUP:
			if (argc == 0) {
				snprintf(tmp, PATH_MAX, "%s", PkgMgr.wrkdir);
			}
			else if (isdir(*argv)) {
				snprintf(tmp, PATH_MAX, "%s", *argv);
				argv++;
				argc--;
			}
			else {
				snprintf(tmp, PATH_MAX, "%s", PkgMgr.wrkdir);
			}
			msg_verbose("Creating backup packages in %s\n", tmp);
			if (!argc) {
				msg_verbose("Creating backup packages for all installed"
					       "packages.  This could take some time.\n");
				PLIST_FOREACH(pe, &PkgMgr) {
					err += pkg_mgr_create_backup(pe->dbdir, tmp);
				}
			}
			else {
				while (*argv) {
					err += pkg_mgr_create_backup(*argv, tmp);
					argv++;
				}
			}
			if (err != 0) {
				msg_warn("%d package creation(s) failed", err);
				return err;
			}
			break;

		case NORMAL:
			if (argc != 1) {
				msg_warn("only one package name can be specified for creatation");
				usage();
			} 
			if (pkg_mgr_create(&pi, &dl, *argv)) {
				msg_warn("Package creation failed");
				return FAIL;
			}
			break;
	}
	return SUCCESS;
}

static void
usage()
{
	fprintf(stderr, "%s\n%s\n%s\n%s\n%s\n%s\n",
"usage: pkg_create [-YNOBhv] [-P pkgs] [-p prefix] [-f contents] [-i iscript]",
"				  [-I piscript] [-k dscript] [-K pdscript] [-r rscript] ",
"				  [-t template] [-X excludefile] [-D displayfile] ",
"				  [-L libdep:pkg] [-R rundep:pkg] [-m mtreefile] ",
"				  [-o origin] -c comment -d description -f packlist ",
"				  pkg-name");
	exit(1);
}

void
cleanup(int sig)
{
	static int in_cleanup = 0;

	if (!in_cleanup) {
		in_cleanup = 1;
			printf("Signal %d received, cleaning up..\n", sig);
	}
	if (sig)
		exit(1);
}
