#ifndef lint
static const char rcsid[] =
  "$FreeBSD: src/usr.sbin/pkg_install/delete/main.c,v 1.21 2001/02/27 09:00:18 sobomax Exp $";
#endif

/*
 *
 * FreeBSD install - a package for the installation and maintainance
 * of non-core utilities.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 * Jordan K. Hubbard
 * 18 July 1993
 *
 * This is the delete module.
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <err.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "list.h"
#include "lib.h"

static char Options[] = "hvDdnfp:";

static void usage __P((void));
void cleanup(int);

int
main(int argc, char **argv)
{
	int ch, err = 0;

	signal(SIGINT, cleanup);
	signal(SIGHUP, cleanup);

	pkg_mgr_init();
	while ((ch = getopt(argc, argv, Options)) != -1)
		switch(ch) {
		case 'v':
			PkgMgr.verbose = TRUE;
			break;

		case 'f':
			PkgMgr.force = TRUE;
			break;

		case 'p':
			strlcpy(PkgMgr.prefix_override, optarg, PATH_MAX);
			break;

		case 'D':
			PkgMgr.noinstall = TRUE;
			break;

		case 'd':
			PkgMgr.cleandirs = TRUE;
			break;

		case 'n':
			PkgMgr.fake = TRUE;
			PkgMgr.verbose = TRUE;
			break;

		case 'h':
		case '?':
		default:
			usage();
			break;
		}

	argc -= optind;
	argv += optind;

	if (!argc) {
		msg_warn("missing package name(s)");
		usage();
	}
	while (*argv) {
		err += pkg_mgr_delete(*argv);
		argv++;
	}
	if (err != 0) {
		msg_warn("%d package deletion(s) failed", err);
		return err;
	}
	return SUCCESS;
}

static void
usage()
{
	fprintf(stderr, "usage: pkg_delete [-vDdnf] [-p prefix] pkg-name ...\n");
	exit(1);
}

void
cleanup(int sig)
{
	static int in_cleanup = 0;

	if (!in_cleanup) {
		in_cleanup = 1;
			printf("Signal %d received, cleaning up..\n", sig);
	}
	if (sig)
		exit(1);
}
