/*
 * XXX: (reg) insert final project name here.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Jeremy D. Lea
 * XXX: (reg) Insert date
 *
 * Based on pkg_install, by:
 *   Jordan K. Hubbard
 *   18 July 1993
 *
 * This header defines the API for a package manager, along with various
 * other functions.  XXX: (reg) Split this into a private and a public
 * header.
 */

#ifndef _INST_LIB_LIB_H_
#define _INST_LIB_LIB_H_

/* Macros */
#define	SUCCESS		(0)
#define	FAIL		(-1)
typedef int error_code;
typedef int error_count;

#ifndef	TRUE
#define	TRUE		(1)
#endif
#ifndef	FALSE
#define	FALSE		(0)
#endif
typedef unsigned int boolean;

/* Where we put logging information by default, else ${PKG_DBDIR} if set */
#define	DEF_LOG_DIR		"/var/db/pkg"
/* just in case we change the environment variable name */
#define	PKG_DBDIR		"PKG_DBDIR"
/* macro to get name of directory where we put logging information */
#define LOG_DIR			(getenv(PKG_DBDIR) ? getenv(PKG_DBDIR) : DEF_LOG_DIR)
#define PKGDB_FILE		"file.db"	/* indexed by filename */

/* Where the ports lives by default */
#define DEF_PORTS_DIR	"/usr/ports"
/* just in case we change the environment variable name */
#define PORTSDIR		"PORTSDIR"
/* macro to get name of directory where we put logging information */
#define PORTS_DIR		(getenv(PORTSDIR) ? getenv(PORTSDIR) : DEF_PORTS_DIR)

/* The name of the package path environment variable */
#define	PKG_PATH		"PKG_PATH"
/* The base URL for fetching remote packages */
#define	PKG_SITE		"PKG_SITE"
/* The name of the "prefix" environment variable given to scripts */
#define	PKG_PREFIX		"PKG_PREFIX"
/* Some temporary space */
#define	PKG_TMPDIR		"PKG_TMPDIR"
#define	SYS_TMPDIR		"TMPDIR"

/* The names of our "special" files */
#define	CONTENTS_FNAME			"+CONTENTS"
#define	COMMENT_FNAME			"+COMMENT"
#define	DESC_FNAME				"+DESC"
#define	INSTALL_FNAME			"+INSTALL"
#define	POST_INSTALL_FNAME		"+POST-INSTALL"
#define	DEINSTALL_FNAME			"+DEINSTALL"
#define	POST_DEINSTALL_FNAME	"+POST-DEINSTALL"
#define	REQUIRE_FNAME			"+REQUIRE"
#define	REQUIRED_BY_FNAME		"+REQUIRED_BY"
#define	DISPLAY_FNAME			"+DISPLAY"
#define	MTREE_FNAME				"+MTREE_DIRS"
#define SPECIAL_FNAME \
	CONTENTS_FNAME " " COMMENT_FNAME " " DESC_FNAME

#define	CMD_CHAR				'@'			/* prefix for extended PLIST cmd */
#define	PKG_SUFFIX				".tgz"		/* The suffix we use for package files */

/*
 * Local type definitions.  These rely on the inclusion of our special
 * list.h to supply the PLIST macros.
 */

/*
 * Structures for a 'which list', which is used to store a list of files,
 * and which packages installed them.  The enum defines whether the file is
 * an ordinary file, a library or an executable.
 *
 * A which list is used in a number of contexts to determine which package
 * installed a particular file.
 */
enum which_t {
	WHICH_FILE, WHICH_LIBRARY, WHICH_EXEC
};

struct _which_entry {
	char file[PATH_MAX];
	char pkg[PATH_MAX];
	enum which_t type;
	PLIST_ENTRY(_which_entry);
};
typedef struct _which_entry which_entry;

struct _which_list {
	PLIST_HEAD(_which_entry);
};
typedef struct _which_list which_list;

/*
 * Similar to a 'which list' a 'depend list' stores a list of dependencies
 * from one package to another.  The enum defines the type of dependency.
 *
 * A depend list has a which list piggy backed onto it, which is used to
 * determine the actual file involved in the dependency, and the actual
 * package which installed it.
 */
enum _depend_t {
	DEPEND_PKG, DEPEND_LIB, DEPEND_RUN
};
typedef enum _depend_t depend_t;

struct _depend_entry {
	char file[PATH_MAX];
	char pkg[PATH_MAX];
	which_entry *w;
	depend_t type;
	PLIST_ENTRY(_depend_entry);
};
typedef struct _depend_entry depend_entry;

struct _depend_list {
	which_list wl;
	PLIST_HEAD(_depend_entry);
};
typedef struct _depend_list depend_list;

/*
 * A require list is used to store a list of packages which require the
 * package in question.
 */
struct _require_entry {
	char pkg[PATH_MAX];
	PLIST_ENTRY(_require_entry);
};
typedef struct _require_entry require_entry;

struct _require_list {
	PLIST_HEAD(_require_entry);
};
typedef struct _require_list require_list;

/*
 * The classic packaging list, which is used to store a representation of
 * the package.  Each plist_entry normally equates to one line in the
 * pkg-plist/+CONTENTS file.  It is very important that this list remain
 * ordered.
 *
 * The name[] stores the text of the line, the md5[] refers to the MD5
 * checksum of the file or link.  The enum stores the type of entry.  The
 * booleans store whether the entry should be ignored in processing, and
 * whether it was merged in from an older PLIST during an update.
 */
enum _plist_t {
	PLIST_INVALID = 0,
	PLIST_FILE, PLIST_CWD, PLIST_CMD, PLIST_CHMOD,
	PLIST_CHOWN, PLIST_CHGRP, PLIST_COMMENT, PLIST_IGNORE,
	PLIST_NAME, PLIST_UNEXEC, PLIST_SRC, PLIST_DISPLAY,
	PLIST_PKGDEP, PLIST_MTREE, PLIST_DIR_RM, PLIST_IGNORE_INST,
	PLIST_MD5, PLIST_ORIGIN, PLIST_LIBDEP, PLIST_RUNDEP,
	PLIST_OPTION_INPLACE, PLIST_OPTION_PRESERVE
};
typedef enum _plist_t plist_t;

struct _plist_entry {
	char *name;
	char *md5;
	plist_t type;
	boolean ignore;
	boolean	merged;
	PLIST_ENTRY(_plist_entry);
};
typedef struct _plist_entry plist_entry;

struct _plist {
	PLIST_HEAD(_plist_entry);
};
typedef struct _plist plist;

#define PKG_OPTION_NONE		0x0000	/* No options on the package. */
#define PKG_OPTION_INPLACE	0x0001	/* The package should be extracted
									   in place. */
#define PKG_OPTION_PRESERVE	0x0002	/* We should try to preserve the
									   installed system for later
									   restoration, should the package
									   install fail. */
/*
 * These are pointers to all of the special files.
 */
struct _pkg_info {
	char *contents;
	char *comment;
	char *desc;
	char *display;
	char *install;
	char *deinstall;
	char *require;
	char *prefix;
	char *srcdir;
	char *mtree;
	char *origin;
};
typedef struct _pkg_info pkg_info;

#define PKG_INFO_INITIALIZER \
	NULL, NULL, NULL, NULL, NULL, NULL, \
	NULL, NULL, NULL, NULL, NULL

/*
 * A package is the top level structure for any particular package and
 * contains all of the information needed to manipulate the package whether
 * it is installed or not.
 */
struct _package {
	char name[PATH_MAX];
	char origin[PATH_MAX];
	char display[PATH_MAX];
	char mtree[PATH_MAX];
	unsigned int options;
	plist plist;
	depend_list depends;
	require_list requires;
};
typedef struct _package package;

/*
 * From here on we deal with structures which only apply to installed
 * packages.
 *
 * The package manager is a state machine, and this enum defines the states. 
 * Generally a package will move through this list from initial installation
 * to final deletion.
 */
typedef enum {
	PKG_LEAVE = -2,				/* We want to leave behind tmpdir */
	PKG_FAIL = -1,				/* Something went wrong */
	PKG_UNKNOWN,				/* We don't know anything about this pkg */
	PKG_FETCH,					/* We're trying to fetch it */
	PKG_SPECIAL,				/* We've extracted the special files only (into tmpdir) */
	PKG_EXTRACTED,				/* We've extracted the entire package */
	PKG_TEMP,					/* We're just looking today */
	PKG_INSTALL,				/* We're in the process of installing */
	PKG_INSTALLED,				/* We've installed, but it's not registered yet */
	PKG_REGISTERED,				/* We've got it installed */
	PKG_CHANGED,				/* Someone changed it's depends */
	PKG_DELETE,					/* We're trying to delete it */
} pkg_status_t;

/*
 * A package entry describes the system dependant part of the package: Where
 * it's coming from, has been extracted, and it's final resting place.
 */
struct _pkg_entry {
	char url[PATH_MAX];
	char tmpdir[PATH_MAX];
	char dbdir[PATH_MAX];
	package *pkg;
	pkg_status_t status;
	PLIST_ENTRY(_pkg_entry);
};
typedef struct _pkg_entry pkg_entry;

/*
 * A simple structure for determining a list of packages some search
 * criteria, so that they can be manipulated or interogated.
 */
struct _match_entry {
	pkg_entry *pe;
	PLIST_ENTRY(_match_entry);
};
typedef struct _match_entry match_entry;

struct _match_list {
	PLIST_HEAD(_match_entry);
};
typedef struct _match_list match_list;


/*
 * Finally we come to the package manager.  There is only ever one of these,
 * which will lock the package database on creation, and only release it
 * when it is destroyed.  It's operation relies on having total control over
 * the database.
 *
 * This structure forms the core of managing a list of package entries, and
 * also stores all of the global options which have been defined.  Most
 * interesting are the three callback functions, which can be used to
 * interact with the user.  The package manager *always* returns from
 * whatever operation it was asked to perform, but might call these
 * functions if there was an error.
 */
struct _pkg_mgr {
	char wrkdir[PATH_MAX];
	char dbdir[PATH_MAX];
	char mktmp_template[PATH_MAX];
	char prefix_override[PATH_MAX];
	boolean	verbose;
	boolean fake;
	boolean force;
	boolean noinstall;
	boolean norecord;
	boolean cleandirs;
	boolean dereference;
	char * excludefrom;
	char * infoprefix;

	void (*msg)(const char *, va_list);
	void (*warn)(const char *, va_list);
	void (*err)(int, const char *, va_list);
	PLIST_HEAD(_pkg_entry);
};
typedef struct _pkg_mgr pkg_mgr;
extern pkg_mgr PkgMgr;

/*
 * A simple structure for an index entry.  There is only one global
 * index, managed by the package manager.  It is only loaded if needed.
 */
struct _index_entry {
    char name[PATH_MAX];
    char origin[PATH_MAX];
	PLIST_ENTRY(_index_entry);
};
typedef struct _index_entry index_entry;

struct _index_list {
    char index[PATH_MAX];
    char mask[7];
	PLIST_HEAD(_index_entry);
};
typedef struct _index_list index_list;
extern index_list PkgIndex;

/* Prototypes */
/* exec.c */
error_code			format_cmd(char *, int, const char *, const char *, const char *);
error_code			vsystem(const char *, ...);
char				*vpipe(const char *, ...);

/* str.c */
const char			*basename_of(const char *);
char				*pkgname_of(char *, const char *, int);
char				*pkgroot_of(char *, const char *, int);
const char			*version_of(const char *, int *, int *);
int					version_cmp(const char *, const char *);
char				*copy_string(const char *);

/* file.c */
boolean				isnode(const char *);
boolean				isdir(const char *);
boolean				isemptydir(const char *);
boolean				isemptyfile(const char *);
boolean				isfile(const char *);
boolean				isempty(const char *);
boolean				issymlink(const char *);
boolean				isURL(const char *);
char				*find_file(const char *, const char *);
char				*read_line(char *, int, FILE *);
char				*read_file(const char *);
error_code			write_file(const char *, const char *);
error_code			copy_file(const char *, const char *, const char *);
error_code			move_file(const char *, const char *, const char *);
error_code			link_file(const char *, const char *);
error_code			apply_perms(const char *, const char *, const char *, const char *, const char *);
error_code			make_hierarchy(const char *, const char *, const char *, const char *);
error_code			delete_hierarchy(const char *, boolean, boolean);
off_t				min_free(const char *);
error_code			make_tmpdir(char *, size_t);
error_code 			enter_dir(char *);
error_code			remove_dir(const char *);
char				*make_md5(const char *);
boolean				check_md5(const char *, const char *);

/* msg.c */
void				msg_note(const char *, ...);
void				msg_verbose(const char *, ...);
void				msg_warn(const char *, ...);
void				msg_error(int, const char *, ...);

/* filedb.c */
error_code			add_package_filedb(package *);
error_code			remove_package_filedb(package *);
error_code			which_filedb(which_entry *w);

/* which.c */
which_entry			*new_which_entry(void);
void				free_which(which_list *);
which_entry			*add_which_entry(which_list *, enum which_t, char *);
error_code			add_which_plist(which_list *, char *);
boolean				check_which(which_entry *, package *);
boolean				scan_which_entry(which_entry *);
boolean				scan_which(which_list *);

/* depend.c */
depend_entry		*new_depend_entry(void);
void				free_depend(depend_list *);
depend_entry		*add_depend_entry(depend_list *, depend_t, char *, char *);
boolean				check_depend(depend_entry *);
require_entry		*new_require_entry(void);
void				free_require(require_list *);
require_entry		*add_require_entry(require_list *, char *);
error_code			delete_require_entry(require_list *, char *);

/* plist.c */
plist_entry			*new_plist_entry(void);
void				free_plist(plist *);
plist_entry			*find_plist_entry(plist *, plist_t, char *);
plist_entry			*add_plist_entry(plist *, plist_t, char *);
plist_entry			*add_plist_top(plist *, plist_t, char *);
error_code			delete_plist_entry(plist *, plist_entry *);
plist_t				read_plist_entry(char *, char **);
error_code			write_plist_entry(FILE *, plist_t, char *);
error_code			move_plist_entry(plist *, plist_entry *, plist *);

/* show.c */
#define	PKG_SHOW_NONE		0x0000
#define	PKG_SHOW_COMMENT	0x0001
#define PKG_SHOW_DESC		0x0002
#define PKG_SHOW_PLIST		0x0004
#define PKG_SHOW_INSTALL	0x0008
#define PKG_SHOW_DEINSTALL	0x0010
#define PKG_SHOW_REQUIRE	0x0020
#define PKG_SHOW_PREFIX		0x0040
#define PKG_SHOW_INDEX		0x0080
#define PKG_SHOW_FILES		0x0100
#define PKG_SHOW_DISPLAY	0x0200
#define PKG_SHOW_REQBY		0x0400
#define PKG_SHOW_MTREE		0x0800
#define PKG_SHOW_SIZE		0x1000
#define PKG_SHOW_ORIGIN		0x2000
#define PKG_SHOW_CHKSUM		0x4000
#define PKG_SHOW_WHICH		0x8000

error_code			show_file(char *, char *);
error_code			show_plist(char *, package *, plist_t);
error_code			show_files(char *, package *);
error_code			show_depend(char *, depend_list *);
error_code			show_index(char *, char *);
error_code			show_size(char *, package *);
error_code			show_chksum(char *, package *);
error_code			show_origin(char *, package *);
error_code			show_which(package *, which_list *);
error_code			show_version(const char *, const char *, const char *);

/* package.c */
package				*new_package(void);
void				free_package(package *);
error_code			fetch_package(char *);
error_code			unpack_package(char *, char *);
package				*create_package(pkg_info *, depend_list *, char *);
error_code			read_package(package *);
error_code			write_package(package *);
error_code			extract_package(package *);
error_code			install_package(package *);
error_code			register_package(package *, char *);
error_code			depend_package(package *, char*);
error_code			undepend_package(package *);
boolean				required_package(package *);
error_code			delete_package(package *);
error_code 			make_package(package *, char *);
error_code			show_package(package *, unsigned int);
error_code			upgrade_package(package *, package *);
error_code			version_package(package *);

/* entry.c */
pkg_entry			*new_pkg_entry(void);
void				cleanup_pkg_entry(pkg_entry *);
pkg_entry			*add_pkg_entry(void);
pkg_entry			*find_pkg_entry(const char *);
error_code			match_pkg_entry(match_list *, const char *);
error_code			fetch_pkg_entry(pkg_entry *, char *);
error_code			extract_pkg_entry(pkg_entry *, boolean);
error_code			scan_pkg_entry(pkg_entry *);
error_code			install_pkg_entry(pkg_entry *);
error_code			register_pkg_entry(pkg_entry *);
error_code			write_pkg_entry(pkg_entry *);
error_code			delete_pkg_entry(pkg_entry *);
pkg_entry			*create_pkg_entry(pkg_info *, depend_list *, char *);
error_code			package_pkg_entry(pkg_entry *, char *);
error_code			show_pkg_entry(pkg_entry *, unsigned int, which_list *);
error_code			upgrade_pkg_entry(pkg_entry *, pkg_entry *);
error_code			version_pkg_entry(pkg_entry *);

/* manager.c */
error_code			pkg_mgr_init(void);
void				pkg_mgr_cleanup(void);
boolean				pkg_mgr_check(char *);
error_code			pkg_mgr_add_latest(char *);
error_code			pkg_mgr_add(char *);
error_code			pkg_mgr_add_master(char *, char *);
error_code			pkg_mgr_add_slave(char *);
error_code			pkg_mgr_add_port(char *);
error_code			pkg_mgr_delete(char *);
error_code			pkg_mgr_create(pkg_info *, depend_list *, char *);
error_code			pkg_mgr_create_port(pkg_info *, depend_list *, char *, char *);
error_code			pkg_mgr_create_backup(char *, char *);
error_code			pkg_mgr_show(char *, unsigned int, which_list *);
error_count			pkg_mgr_version(const char *);
void				pkg_mgr_index(void);

#endif /* _INST_LIB_LIB_H_ */
