/*
 * XXX: (reg) insert final project name here.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Jeremy D. Lea
 * XXX: (reg) Insert date
 *
 * Based on FreeBSD's <sys/queue.h>.
 *
 * Macros for managing a singly linked list, with a head which can contain
 * real data, and a tail pointer.
 */

#ifndef _INST_LIB_LIST_H_
#define	_INST_LIB_LIST_H_

#define	PLIST_HEAD(type) \
	struct type *pl_head;	/* first element */ \
	struct type *pl_last;	/* last element */

#define	PLIST_HEAD_INITIALIZER	NULL, NULL

#define	PLIST_ENTRY(type) \
	struct type *pl_next;	/* next element */
 
/*
 * List access functions.
 */
#define	PLIST_EMPTY(head)	((head)->pl_head == NULL)

#define	PLIST_FIRST(head)	((head)->pl_head)

#define	PLIST_LAST(head)	((head)->pl_last)

#define	PLIST_NEXT(elm)		((elm)->pl_next)

#define	PLIST_FOREACH(var, head) \
	for ((var) = PLIST_FIRST((head)); \
	    (var); \
	    (var) = ( (var) ? PLIST_NEXT((var)) : NULL ) )

#define	PLIST_INIT(head) do { \
	PLIST_FIRST((head)) = NULL; \
	PLIST_LAST((head)) = NULL; \
} while (0)

#define	PLIST_INSERT_HEAD(head, elm) do { \
	if (PLIST_FIRST((head)) == NULL) \
		PLIST_LAST((head)) = (elm); \
	PLIST_NEXT((elm)) = PLIST_FIRST((head)); \
	PLIST_FIRST((head)) = (elm); \
} while (0)

#define	PLIST_INSERT_AFTER(head, plistelm, elm) do { \
	if (PLIST_EMPTY((head))) \
		PLIST_INSERT_HEAD(head, elm); \
	else { \
		PLIST_NEXT((elm)) = PLIST_NEXT((plistelm)); \
		PLIST_NEXT((plistelm)) = (elm);	\
		if ((plistelm) == PLIST_LAST((head))) \
			PLIST_LAST((head)) = (elm); \
	} \
} while (0)

#define	PLIST_INSERT_TAIL(head, elm) do { \
	PLIST_INSERT_AFTER(head, PLIST_LAST((head)), elm); \
} while (0)

/*
 * The value of tmp is overwritten with the old head, or NULL if
 * the list is empty.
 */
#define	PLIST_REMOVE_HEAD(head, tmp) do { \
	(tmp) = PLIST_FIRST((head)); \
	if ((tmp) != NULL) { \
		PLIST_FIRST((head)) = PLIST_NEXT((tmp)); \
		if (PLIST_LAST((head)) == (tmp)) \
			PLIST_LAST((head)) = NULL; \
	} \
} while (0)

/*
 * The value of tmp is overwritten, with a valid current element or
 * NULL if elm was not in the list.
 */
#define	PLIST_REMOVE(head, elm, tmp) do { \
	if (PLIST_FIRST((head)) == (elm)) { \
		PLIST_REMOVE_HEAD((head), elm); \
		(tmp) = PLIST_FIRST((head)); \
	} \
	else if ((tmp) != NULL && PLIST_NEXT((tmp)) == (elm)) { \
		PLIST_NEXT((tmp)) = PLIST_NEXT((elm)); \
	} \
	else { \
		(tmp) = PLIST_FIRST((head)); \
		while ((tmp) != NULL && PLIST_NEXT((tmp)) != (elm)) \
			(tmp) = PLIST_NEXT((tmp)); \
		if ((tmp) != NULL) \
			PLIST_NEXT((tmp)) = PLIST_NEXT((elm)); \
	} \
	if (PLIST_LAST((head)) == (elm)) \
		PLIST_LAST((head)) = (tmp); \
} while (0)

#endif /* !_INST_LIB_LIST_H_ */
