/*
 * XXX: (reg) insert final project name here.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Jeremy D. Lea
 * XXX: (reg) Insert date
 *
 * Based on pkg_install, by:
 *   Jordan K. Hubbard
 *   18 July 1993
 *
 * Package manager.
 */

#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <md5.h>
#include <fts.h>
#include <fetch.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <sys/utsname.h>
#include "list.h"
#include "lib.h"

int getosreldate(void);
int	fname_cmp(const FTSENT **, const FTSENT **);

index_list PkgIndex;
pkg_mgr PkgMgr;

/*
 * Initialise the Package Manager.  This must be called before any of the
 * other functions.  pkg_mgr_cleanup is registered with atexit(), so that
 * you don't have to make sure you call it in all corner cases.
 */
error_code
pkg_mgr_init(void)
{
	pkg_entry *pe;
	FTS *ftsp;
	FTSENT *f;
	char *paths[2];
	struct stat stat_s;
    
	/* Set a reasonable umask */
	umask(022);

	/* Make sure that we clean up after ourselves */
	atexit(pkg_mgr_cleanup);

	/* Setup our defaults and callbacks */
	PkgMgr.excludefrom = "";
	PkgMgr.infoprefix = "";
	PkgMgr.verbose = FALSE;
	PkgMgr.fake = FALSE;
	PkgMgr.force = FALSE;
	PkgMgr.noinstall = FALSE;
	PkgMgr.norecord = FALSE;
	PkgMgr.cleandirs = FALSE;
	PkgMgr.dereference = FALSE;
	PkgMgr.msg = &vprintf;
	PkgMgr.warn = &vwarnx;
	PkgMgr.err = &verrx;
	PLIST_INIT(&PkgMgr);

	PkgMgr.wrkdir[0] = '\0';
	if (!getcwd(PkgMgr.wrkdir, PATH_MAX))
		msg_warn("getcwd() failed!");
	strlcpy(PkgMgr.dbdir, LOG_DIR, PATH_MAX);
	if (!isdir(PkgMgr.dbdir)) {
		msg_warn("Unable to determine package database directoy.  %s is not a directory!  Please set %s to the correct path, or create this directory.", PkgMgr.dbdir, PKG_DBDIR);
		return FAIL;
	}
	(void) stat(PkgMgr.dbdir, &stat_s);
	if (!PkgMgr.fake && getuid() && geteuid() != stat_s.st_uid) {
		if (!PkgMgr.force) {
			msg_error(2, "you do not own %s, use -f to force", PkgMgr.dbdir);
			return FAIL;
		} else
			msg_warn("you do not own %s (proceeding anyway)", PkgMgr.dbdir);
	}
	strcpy(PkgMgr.mktmp_template, "instmp.XXXXXX");
	strcpy(PkgMgr.prefix_override, "");

	/* Loop through the dbdir looking for registered packages */
	paths[0] = PkgMgr.dbdir;
	paths[1] = NULL;
	ftsp = fts_open(paths, FTS_LOGICAL | FTS_NOCHDIR | FTS_NOSTAT, fname_cmp);
	if (ftsp != NULL) {
		while ((f = fts_read(ftsp)) != NULL) {
			if (f->fts_info == FTS_D && f->fts_level == 1) {
				if ((pe = add_pkg_entry()) == NULL)
					return FAIL;
				snprintf(pe->dbdir, PATH_MAX, "%s", f->fts_path);
				pe->status = PKG_REGISTERED;
			}
		}
		fts_close(ftsp);
	}
	
	/* Take care of the index.  This is demand loaded later. */
	snprintf(PkgIndex.index, PATH_MAX, "%s/INDEX", PORTS_DIR);
	strcpy(PkgIndex.mask, "<=>*?!");
	PLIST_INIT(&PkgIndex);
	
	return SUCCESS;
}

/*
 * A simple check of if a package is installed.
 */
boolean
pkg_mgr_check(char *pkgname)
{
	pkg_entry *pe;
	
	if ((pe = find_pkg_entry(pkgname)) == NULL)
		return FALSE;
	if (pe->status != PKG_REGISTERED)
		return FALSE;
	return TRUE;
}

/*
 * Add a package, from the LATEST_LINK collection.  XXX: (reg) This needs to
 * be updated.
 */
error_code
pkg_mgr_add_latest(char *pkgarg)
{
	const char *packageroot = "ftp://ftp.FreeBSD.org/pub/FreeBSD/ports/";
	static char sitepath[PATH_MAX];
	int reldate;
	struct utsname u;

	if (pkgarg == NULL)
		return FAIL;
	if (getenv(PKG_SITE))
		strlcpy(sitepath, getenv(PKG_SITE), PATH_MAX);
	else {
		strlcpy(sitepath, packageroot, PATH_MAX);
		uname(&u);
		strlcat(sitepath, u.machine, PATH_MAX);
		reldate = getosreldate();
		if (reldate >= 500000)
			strlcat(sitepath, "/packages-current/Latest/", PATH_MAX);
	}
	strlcat(sitepath, pkgarg, PATH_MAX);

	return pkg_mgr_add(sitepath);
}

/*
 * Add a package, from either a file or URL.
 */
error_code
pkg_mgr_add(char *pkgarg)
{
	pkg_entry *pe;
	
	if (pkgarg == NULL)
		return FAIL;
	if ((pe = add_pkg_entry()) == NULL)
		return FAIL;
	if (fetch_pkg_entry(pe, pkgarg) == FAIL)
		return FAIL;
	if (install_pkg_entry(pe) == FAIL)
		return FAIL;
	return SUCCESS;
}

/*
 * Just extract the port into a temporary directory.
 */ 
error_code
pkg_mgr_add_master(char *pkgarg, char *pkgtmp)
{
	pkg_entry *pe;
	
	if (pkgarg == NULL || pkgtmp == NULL)
		return FAIL;
	if ((pe = add_pkg_entry()) == NULL)
		return FAIL;
	if (fetch_pkg_entry(pe, pkgarg) == FAIL)
		return FAIL;
	if (extract_pkg_entry(pe, FALSE) == FAIL)
		return FAIL;
	snprintf(pkgtmp, PATH_MAX, pe->tmpdir);
	pe->status = PKG_LEAVE;
	return FALSE;
}

/*
 * Add a package in slave mode, from the given tmpdir.
 */
error_code
pkg_mgr_add_slave(char *pkgtmp)
{
	pkg_entry *pe;
	
	if (pkgtmp == NULL)
		return FAIL;
	if ((pe = add_pkg_entry()) == NULL)
		return FAIL;
	strlcpy(pe->tmpdir, pkgtmp, PATH_MAX);
	pe->status = PKG_EXTRACTED;
	if (install_pkg_entry(pe) == FAIL)
		return FAIL;
	return SUCCESS;
}

/*
 * Add a package from a port.  This should have been created for us by
 * pkg_mgr_create_port().
 */
error_code
pkg_mgr_add_port(char *pkgtmp)
{
	pkg_entry *pe;
	
	if (pkgtmp == NULL)
		return FAIL;
	if ((pe = add_pkg_entry()) == NULL)
		return FAIL;
	strlcpy(pe->tmpdir, pkgtmp, PATH_MAX);
	pe->status = PKG_INSTALLED;
	if (register_pkg_entry(pe) == FAIL)
		return FAIL;
	return SUCCESS;
}

/*
 * Delete all installed packages which match the argument.
 */
error_code
pkg_mgr_delete(char *pkgarg)
{
	boolean err = FALSE;
	match_list ml =  { NULL };
	match_entry *m;

	if (pkgarg == NULL)
		return FAIL;
	if (match_pkg_entry(&ml, pkgarg))
		return FAIL;
	PLIST_FOREACH(m, &ml) {
		if (delete_pkg_entry(m->pe) == FAIL) {
			err = TRUE;
			break;
		}
	}
	while (!PLIST_EMPTY(&ml)) {
		PLIST_REMOVE_HEAD(&ml, m);
		free(m);
	}
	if (err)
		return FAIL;
	return SUCCESS;
}

/*
 * Create a port from installed files, and supplied information.  It does
 * not matter if the package is registered or not.
 */
error_code
pkg_mgr_create(pkg_info *pi, depend_list *dl, char *pkgURL)
{
	pkg_entry *pe;
	
	if (pi == NULL || dl == NULL || pkgURL == NULL)
		return FAIL;
	if ((pe = create_pkg_entry(pi, dl, pkgURL)) == NULL)
		return FAIL;
	/* If this is not true, we'll crash later... */
	pe->status = PKG_INSTALLED;
	if (package_pkg_entry(pe, pkgURL) == FAIL)
		return FAIL;
	return SUCCESS;
}

/*
 * Fake creating a package from the supplied information, so ports can be
 * installed.
 */
error_code
pkg_mgr_create_port(pkg_info *pi, depend_list *dl, char *pkgURL, char *pkgtmp)
{
	pkg_entry *pe;
	
	if (pi == NULL || dl == NULL || pkgURL == NULL || pkgtmp == NULL)
		return FAIL;
	if ((pe = create_pkg_entry(pi, dl, pkgURL)) == NULL)
		return FAIL;
	strlcpy(pkgtmp, pe->tmpdir, PATH_MAX);
	pe->status = PKG_LEAVE;
	return SUCCESS;
}

/*
 * Create a package from an installed package
 */
error_code
pkg_mgr_create_backup(char *pkgname, char *backupdir)
{
	pkg_entry *pe;
	
	if (pkgname == NULL || backupdir == NULL)
		return FAIL;
	if ((pe = find_pkg_entry(pkgname)) == NULL)
		return FAIL;
	if (pe->status != PKG_REGISTERED)
		return FAIL;
	if (package_pkg_entry(pe, backupdir) == FAIL)
		return FAIL;
	return SUCCESS;
}

/*
 * Show information about all the matching packages, fetching and extracting
 * it if it is a URL.
 */
error_code
pkg_mgr_show(char *pkgarg, unsigned int flags, which_list *wl)
{
	char *tmp;
	boolean err = FALSE;
	match_list ml =  { NULL };
	match_entry *m;
	pkg_entry *pe;

	if (pkgarg == NULL || wl == NULL)
		return FAIL;
	if ((tmp = find_file(NULL, pkgarg)) != NULL) {
		free(tmp);
		if ((pe = add_pkg_entry()) == NULL)
			return FAIL;
		if (fetch_pkg_entry(pe, pkgarg) == FAIL)
			return FAIL;
		if (extract_pkg_entry(pe, FALSE) == FAIL)
			return FAIL;	
		pe->status = PKG_TEMP;
		/* We can't size packages that are not installed */
		flags &= ~PKG_SHOW_SIZE;
		if (show_pkg_entry(pe, flags, wl) == FAIL)
			return FAIL;
	} else if (match_pkg_entry(&ml, pkgarg) == FAIL)
		return FAIL;
	else {
		PLIST_FOREACH(m, &ml) {
			if (show_pkg_entry(m->pe, flags, wl)) {
				err = TRUE;
				break;
			}
		}
	}
	while (!PLIST_EMPTY(&ml)) {
		PLIST_REMOVE_HEAD(&ml, m);
		free(m);
	}
	if (err)
		return FAIL;
	return SUCCESS;
}

/*
 * Determine if the matching (installed) packages need updating.
 */
error_count
pkg_mgr_version(const char *pkgarg)
{
	error_count err_cnt = 0;
	match_list ml = { NULL };
	match_entry *m;

	if (pkgarg == NULL)
		return FAIL;
	if (match_pkg_entry(&ml, pkgarg) == FAIL)
		return FAIL;
	PLIST_FOREACH(m, &ml)
		err_cnt += version_pkg_entry(m->pe);
	while (!PLIST_EMPTY(&ml)) {
		PLIST_REMOVE_HEAD(&ml, m);
		free(m);
	}
	return err_cnt;
}

/*
 * Load the Index.
 */
void
pkg_mgr_index(void)
{
	char *ch, tmp[PATH_MAX];
	FILE *fp;
	size_t len;
	index_entry *ie;

	/* Try to find and open the INDEX */
	if (isURL(PkgIndex.index))
		fp = fetchGetURL(PkgIndex.index, "");
	else
		fp = fopen(PkgIndex.index, "r");
	if (fp == NULL) {
		msg_error(2, "Unable to open INDEX in %s.", __func__);
		return;
	}
	while ((ch = fgetln(fp, &len)) != NULL) {
		strlcpy(tmp, ch, MIN(len, PATH_MAX));
		/* The INDEX has pkgname|portdir|... */
		if ((ch = strchr(tmp, '|')) != NULL)
			ch[0] = '\0';
		if (ch != NULL && (ch = strchr(&ch[1], '|')) != NULL)
			ch[0] = '\0';
		/* Look backwards for the last two dirs = origin */
		while (ch != NULL && *--ch != '/')
			if (ch[0] == '\0')
				ch = NULL;
		while (ch != NULL && *--ch != '/')
			if (ch[0] == '\0')
				ch = NULL;
		if (ch == NULL) {
			msg_error(2, "%s: The INDEX does not appear to be valid!", __func__);
			return;
		}
		if ((ie = malloc(sizeof(index_entry))) == NULL) {
			msg_error(2, "%s: Unable to allocate memory.", __func__);
			return;
		}
		strlcpy(ie->name, tmp, PATH_MAX);
		strlcpy(ie->origin, &ch[1], PATH_MAX);
		PLIST_INSERT_TAIL(&PkgIndex, ie);
	}
	fclose(fp);
}

/*
 * Orderly destruction of our structures and removal of temporary files and
 * directories.
 */
void
pkg_mgr_cleanup(void)
{
	pkg_entry *pe;
	index_entry *ie;
	static int in_cleanup = 0;

	if (!in_cleanup) {
		in_cleanup = 1;
		while (!PLIST_EMPTY(&PkgMgr)) {
			PLIST_REMOVE_HEAD(&PkgMgr, pe);
			cleanup_pkg_entry(pe);
			free(pe);
		}
		if (strlen(PkgMgr.wrkdir) > 0 && isdir(PkgMgr.wrkdir))
			chdir(PkgMgr.wrkdir);
		PkgMgr.wrkdir[0] = '\0';
		PkgMgr.dbdir[0] = '\0';
		PkgMgr.mktmp_template[0] = '\0';
		PkgMgr.prefix_override[0] = '\0';
		while (!PLIST_EMPTY(&PkgIndex)) {
			PLIST_REMOVE_HEAD(&PkgIndex, ie);
			free(ie);
		}
	}
}

/* 
 * We only need this here...
 */
int
fname_cmp(const FTSENT **a, const FTSENT **b)
{
	return strcmp((*a)->fts_name, (*b)->fts_name);
}
