/*
 * XXX: (reg) insert final project name here.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Jeremy D. Lea
 * XXX: (reg) Insert date
 *
 * Based on pkg_install, by:
 *   Jordan K. Hubbard
 *   18 July 1993
 *
 * Various display routines for the info module.  XXX: (reg) These should be
 * moved out of the library, since they are of limited use to a GUI.  We
 * shouldn't use printf() from the library.  We shouldn't be using
 * PkgMgr.verbose, or using msg_verbose either.
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <stdlib.h>
#include "list.h"
#include "lib.h"

/*
 * Show the contents of one of the special files.
 */
error_code
show_file(char *title, char *fname)
{
	char *cp;

	if (title == NULL || fname == NULL)
		return FAIL;
	msg_verbose("%s%s", PkgMgr.infoprefix, title);
	if ((cp = read_file(fname)) == NULL)
		return FAIL;
	printf("%s\n", cp);
	free(cp);
	return SUCCESS;
}

/*
 * Show the index entry for a package (pkgname and +COMMENT).
 */
error_code
show_index(char *pkgname, char *fname)
{
	char *cp;

	if (pkgname == NULL || fname == NULL)
		return FAIL;
	if ((cp = read_file(fname)) == NULL)
		return FAIL;
	*strchr(cp, '\n') = '\0';
	printf("%s%-19.19s %-59.59s\n", PkgMgr.infoprefix, pkgname, cp);
	return SUCCESS;
}

/*
 * Show a packing list item type.  If type is -1, show all.
 */
error_code
show_plist(char *title, package *pkg, plist_t type)
{
	depend_entry *d;
	plist_entry *p;
	boolean v = !PkgMgr.verbose;

	if (title == NULL || pkg == NULL)
		return FAIL;
	msg_verbose("%s%s", PkgMgr.infoprefix, title);
	/* Warning! Warning! Ugly printf hacks! */
	printf(v ? "@name %s\n" : "\tPackage name: %s\n", pkg->name);
	printf(v ? "@comment ORIGIN:%s\n" : "\tPackage origin: %s\n", pkg->origin);
	printf(v ? "@display %s\n" : "\tInstall message file: %s\n", pkg->display);
	printf(v ? "@mtree %s\n" : "\tPackage mtree file: %s\n", pkg->mtree);
	if (pkg->options & PKG_OPTION_INPLACE)
		printf(v ? "@option extrace-in-place\n" : "\tOption: extract-in-place\n");
	if (pkg->options & PKG_OPTION_PRESERVE)
		printf(v ? "@option perserve\n" : "\tOption: preserve\n");
	PLIST_FOREACH(d, &pkg->depends) {
		printf(v ? "@pkgdep " : "\tDepends on ");
		switch (d->type) {
		case DEPEND_PKG:
			printf(v ? " %s\n" : "package %s\n", d->pkg);
			break;
		case DEPEND_LIB:
			printf(v ? "@comment LIBDEP:%s\n" : "shared library %s supplied by %s", d->file, d->pkg);
			break;
		case DEPEND_RUN:
			printf(v ? "@comment RUNDEP:%s\n" : "executable %s supplied by %s", d->file, d->pkg);
			break;
		}
	}
	PLIST_FOREACH(p, &pkg->plist) {
		if (p->type != type && type != -1) {
			p = PLIST_NEXT(p);
			continue;
		}
		switch(p->type) {
		case PLIST_FILE:
			if (p->ignore)
				printf(v ? "@ignore\n%s\n" : "File: %s (ignored)\n", p->name);
			else
				printf(v ? "%s\n" : "File: %s\n", p->name);
			break;

		case PLIST_CWD:
			printf(v ? "@cwd %s\n" : "\tCWD to %s\n", p->name);
			break;

		case PLIST_SRC:
			printf(v ? "@srcdir %s\n" : "\tSRCDIR to %s\n", p->name);
			break;

		case PLIST_CMD:
			printf(v ? "@exec %s\n" : "\tEXEC '%s'\n", p->name);
			break;

		case PLIST_UNEXEC:
			printf(v ? "@unexec %s\n" : "\tUNEXEC '%s'\n", p->name);
			break;

		case PLIST_CHMOD:
			printf(v ? "@chmod %s\n" : "\tCHMOD to %s\n",
				   p->name ? p->name : "(clear default)");
			break;

		case PLIST_CHOWN:
			printf(v ? "@chown %s\n" : "\tCHOWN to %s\n",
				   p->name ? p->name : "(clear default)");
			break;

		case PLIST_CHGRP:
			printf(v ? "@chgrp %s\n" : "\tCHGRP to %s\n",
				   p->name ? p->name : "(clear default)");
			break;

		case PLIST_COMMENT:
			printf(v ? "@comment %s\n" : "\tComment: %s\n", p->name);
			break;

		case PLIST_DIR_RM:
			printf(v ? "@dirrm %s\n" : "\tDeinstall directory remove: %s\n", p->name);
			break;
		
		case PLIST_NAME:
		case PLIST_ORIGIN:
		case PLIST_DISPLAY:
		case PLIST_MTREE:
		case PLIST_OPTION_INPLACE:
		case PLIST_OPTION_PRESERVE:
		case PLIST_IGNORE:
		case PLIST_IGNORE_INST:
		case PLIST_PKGDEP:
		case PLIST_LIBDEP:
		case PLIST_RUNDEP:
			msg_warn("packing list contains nonsense items");
			break;
			
		default:
			msg_error(2, "%s: invalid packing list type", __func__);
			return FAIL;
		}
	}
	return SUCCESS;
}

/*
 * Show all files in the packing list (except ignored ones).
 */
error_code
show_files(char *title, package *pkg)
{
	plist_entry *p;
	char *dir = ".";

	if (title == NULL || pkg == NULL)
		return FAIL;
	msg_verbose("%s%s", PkgMgr.infoprefix, title);
	PLIST_FOREACH(p, &pkg->plist) {
		switch(p->type) {
		case PLIST_FILE:
			if (!p->ignore)
				printf("%s/%s\n", dir, p->name);
			break;

		case PLIST_CWD:
			dir = p->name;
			break;

		default:
			break;
		}
	}
	return SUCCESS;
}

/*
 * Show the dependency information.
 */
error_code
show_depend(char *title, depend_list *dl)
{
	depend_entry *d;

	if (title == NULL || dl == NULL)
		return FAIL;
	msg_verbose("%s%s", PkgMgr.infoprefix, title);
	PLIST_FOREACH(d, dl) {
		if (PkgMgr.verbose) {
			switch(d->type) {
			case DEPEND_PKG:
				msg_verbose("package `%s'\n", d->pkg);
				break;

			case DEPEND_LIB:
				msg_verbose("shared library %s supplied by `%s'\n", d->file, d->pkg);
				break;

			case DEPEND_RUN:
				msg_verbose("executable %s supplied by `%s'\n", d->file, d->pkg);
				break;
			}
		} else
			printf(" %s", d->pkg);
	}
	return SUCCESS;
}

/*
 * Calculate and show size of all installed package files (except ignored
 * ones).
 */
error_code
show_size(char *title, package *pkg)
{
	plist_entry *p;
	char *dir = ".";
	struct stat sb;
	char tmp[PATH_MAX];
	unsigned long size = 0;
	long blksize;
	int headerlen;
	char *descr;

	if (title == NULL || pkg == NULL)
		return FAIL;
	descr = getbsize(&headerlen, &blksize);
	msg_verbose("%s%s", PkgMgr.infoprefix, title);
	PLIST_FOREACH(p, &pkg->plist) {
		switch (p->type) {
		case PLIST_FILE:
			if (p->ignore)
				break;
			snprintf(tmp, PATH_MAX, "%s/%s", dir, p->name);
			if (!lstat(tmp, &sb)) {
				size += sb.st_size;
				msg_verbose("%lu\t%s\n", (unsigned long)howmany(sb.st_size, blksize), tmp);
			}
			break;

		case PLIST_CWD:
			dir = p->name;
			break;

		default:
			break;
		}
	}
	if (PkgMgr.verbose)
		msg_verbose("%lu\t(%s)\n", howmany(size, blksize), descr);
	else
		printf("%lu\n", size);
	return FALSE;
}

/*
 * Show files which don't match the recorded checksum.
 */
error_code
show_chksum(char *title, package *pkg)
{
	plist_entry *p;
	char *dir = ".";
	char tmp[PATH_MAX];

	if (title == NULL || pkg == NULL)
		return FAIL;
	msg_verbose("%s%s", PkgMgr.infoprefix, title);
	PLIST_FOREACH(p, &pkg->plist) {
		switch (p->type) {
		case PLIST_FILE:
			if (p->ignore)
				break;
			if (p->md5) {
				snprintf(tmp, PATH_MAX, "%s/%s", dir, p->name);
				if (!isnode(tmp))
					msg_warn("%s doesn't exist\n", tmp);
				else if (!issymlink(tmp)) {	
					if (!check_md5(tmp, p->md5))
						msg_warn("%s fails original MD5 checksum\n", tmp);
					else
						msg_verbose("%s matched original MD5 checksum\n", tmp);
				}
			}
			break;

		case PLIST_CWD:
			dir = p->name;
			break;

		default:
			break;
		}
	}
	return SUCCESS;
}

/*
 * Show an "origin" path (usually category/portname)
 */
error_code
show_origin(char *title, package *pkg)
{

	if (title == NULL || pkg == NULL)
		return FAIL;
	msg_verbose("%s%s", PkgMgr.infoprefix, title);
	printf("%s\n", pkg->origin);
	return TRUE;
}

/*
 * Show pkgname if we installed said files.
 */
error_code
show_which(package *pkg, which_list *wl)
{
	which_entry *w;
	boolean found = FALSE;

	if (pkg == NULL || wl == NULL)
		return FAIL;
	PLIST_FOREACH(w, wl) {
		if (check_which(w, pkg)) {
			found = TRUE;
			msg_verbose("%s installed by: %s\n", w->file, pkg->name);
		}
	}
	if (found && !PkgMgr.verbose)
		printf("%s\n", pkg->name);
	return SUCCESS;
}

/*
 * Do the work of comparing and outputing. Ugly, but well that's what
 * You get when you try to match perl output in C ;-).
 */
error_code
show_version(const char *installed, const char *latest, const char *source)
{
	char *ch, tmp[PATH_MAX];
	const char *ver;
	int cmp = 0;

	if (!installed || strlen(installed) == 0)
		return FAIL;
	strlcpy(tmp, installed, PATH_MAX);
	if (!PkgMgr.verbose) {
		if ((ch = strrchr(tmp, '-')) != NULL)
			ch[0] = '\0';
	}
	if (latest == NULL) {
		if (source == NULL && strchr(PkgIndex.mask, '!')) {
			printf("%-34s  !", tmp);
			if (PkgMgr.verbose)
				printf("   Comparison failed");
			printf("\n");
		} else if (source != NULL && strchr(PkgIndex.mask, '?')) {
			printf("%-34s  ?", tmp);
			if (PkgMgr.verbose)
				printf("   orphaned: %s", source);
			printf("\n");
		}
	} else if (strchr(latest,'|') != NULL) {
		if (strchr(PkgIndex.mask, '*')) {
			printf("%-34s  *", tmp);
			if (PkgMgr.verbose) {
				strlcpy(tmp, latest, PATH_MAX);
				ch = strchr(tmp, '|');
				ch[0] = '\0';
				ver = version_of(tmp, NULL, NULL);
				printf("   multiple versions (index has %s", ver);
				do {
					ver = version_of(&ch[1], NULL, NULL);
					if ((ch = strchr(&ch[1], '|')) != NULL)
							ch[0] = '\0';
					printf(", %s", ver);
				} while (ch != NULL);
				printf(")");
			}
			printf("\n");
		}
	} else {
		cmp = version_cmp(installed, latest);
		ver = version_of(latest, NULL, NULL);
		if (cmp < 0 && strchr(PkgIndex.mask, '<')) {
			printf("%-34s  <", tmp);
			if (PkgMgr.verbose)
				printf("   needs updating (%s has %s)", source, ver);
			printf("\n");
		} else if (cmp == 0 && strchr(PkgIndex.mask, '=')) {
			printf("%-34s  =", tmp);
			if (PkgMgr.verbose)
				printf("   up-to-date with %s", source);
			printf("\n");
		} else if (cmp > 0 && strchr(PkgIndex.mask, '>')) {
			printf("%-34s  >", tmp);
			if (PkgMgr.verbose)
				printf("   succeeds %s (%s has %s)", source, source, ver);
			printf("\n");
		}
	}
	return SUCCESS;
}
